@@ -1,6 +1,6 @@
 /* Copyright 2007 by Kim Minh Kaplan
  *
- * greyfix.c version 0.3.2
+ * greyfix.c
  *
  * Postfix policy daemon designed to prevent spam using the
  * greylisting method.
@@ -27,6 +27,8 @@
 
 #include <db.h>
 
+#include "policy.h"
+
 /**
  * This determines how many seconds we will block inbound mail that is
  * from a previously unknown (ip, from, to) triplet.  If it is set to
@@ -64,8 +66,6 @@
  * through */
 #define STAT_BLOCKED "blocked"
 
-#define prefixp(s,p) (!strncmp((s),(p),sizeof(p)-1))
-
 struct triplet_data {
     time_t create_time;
     time_t access_time;
@@ -73,36 +73,43 @@ struct triplet_data {
     unsigned long pass_count;
 };
 
-static const char str_smtp[] = "SMTP";
-static const char str_esmtp[] = "ESMTP";
-static const char str_rcpt[] = "RCPT";
-static const char str_action[] = "action=";
-
-static const char *progname;
 static const char *db_home = DEF_DB_HOME;
-static int opt_verbose = 0;
 
 static DB_ENV *dbenv = 0;
 static DB *db = 0;
 static DB *statsdb = 0;
 
-static char *policy_request = 0;
-static size_t policy_request_size = 0;
-static size_t policy_request_fill = 0;
-
 static DBT dbkey = { 0 };
 static DBT dbdata = { 0 };
 static struct triplet_data triplet_data;
 
-static int debug_me = 0;
 static unsigned long greylist_delay = DELAY_MAIL_SECS;
 static unsigned long bloc_max_idle = AUTO_RECORD_LIFE_SECS;
 static unsigned long pass_max_idle = UPDATE_RECORD_LIFE_SECS;
+/* The sequence "%d" is replaced by the number of seconds.  The
+   sequence "%p" is replaced by the singular_string if the delay is
+   one second, by plural_string otherwise.  The sequence "%s" is
+   replaced by the single character " ".  The sequence "%m" is
+   replaced by the receiving domain name.  The sequence "%%" is
+   replaced by the single character "%".  Other sequences beginning
+   with "%" are invalid. */
+static const char *reject_action_fmt = "DEFER_IF_PERMIT Greylisted by "
+    PACKAGE_STRING ", try again in %d second%p."
+    "  See http://www.kim-minh.com/pub/greyfix/ for more information.";
+static const char *greylisted_action_fmt = "PREPEND X-Greyfix: Greylisted by "
+    PACKAGE_STRING " for %d second%p."
+    "  See http://www.kim-minh.com/pub/greyfix/ for more information.";
+static const char *singular_string = "", *plural_string = "s";
 /* As we store IP addresses in Postfix's format, to obtain the network
  address we first strip `ipv4_network_strip_bytes' numbers (between 0
  and 4) then we apply `ipv4_network_mask' on the last byte. */
 static unsigned int ipv4_network_strip_bytes;
 static unsigned int ipv4_network_mask;
+/* --dump-triplets: dump the triplet database to stdout and exit */
+static int dump_triplets;
+
+/* Some Berkeley DB errors are not really error.  Keep them quiet. */
+static int muffle_error = 0;
 
 /**********************************************************************
  * Berkeley DB routines
@@ -110,7 +117,11 @@ static unsigned int ipv4_network_mask;
 static void
 log_db_error(const char *msg, int error)
 {
-    syslog(LOG_ERR, "%s: %s", msg, db_strerror(error));
+    if (muffle_error)
+	syslog(LOG_NOTICE, "NOTICE (this is not an error): %s: %s",
+	       msg, db_strerror(error));
+    else
+	syslog(LOG_ERR, "%s: %s", msg, db_strerror(error));
 }
 
 static void
@@ -129,7 +140,7 @@ prepare_env()
     else {
 	dbenv->set_errcall(dbenv, db_errcall_fcn);
 	rc = dbenv->open(dbenv, db_home,
-			 DB_INIT_CDB | DB_INIT_MPOOL | DB_CREATE, 0);
+			 DB_INIT_LOCK | DB_INIT_MPOOL | DB_CREATE, 0);
 	if (rc)
 	    log_db_error("dbenv->open", rc);
     }
@@ -171,6 +182,9 @@ initialize()
     int rc;
     char *version;
     int major, minor, patch;
+    rc = policy_initialize();
+    if (rc)
+	return rc;
     version = db_version(&major, &minor, &patch);
     if (DB_VERSION_MAJOR != major || DB_VERSION_MINOR != minor) {
 	syslog(LOG_ERR,
@@ -208,20 +222,18 @@ cleanup()
     int rc;
     if (dbkey.data)
 	free(dbkey.data);
-    if (policy_request)
-	free(policy_request);
-    if (statsdb) {
-	rc = statsdb->close(statsdb, 0);
-	statsdb = 0;
-	if (rc)
-	    syslog(LOG_ERR, "BDB-%d: statsdb close: %s", rc, db_strerror(rc));
-    }
     if (db) {
 	rc = db->close(db, 0);
 	db = 0;
 	if (rc)
 	    log_db_error("DB close", rc);
     }
+    if (statsdb) {
+	rc = statsdb->close(statsdb, 0);
+	statsdb = 0;
+	if (rc)
+	    log_db_error("StatsDB close", rc);
+    }
     if (dbenv) {
 	rc = dbenv->close(dbenv, 0);
 	dbenv = 0;
@@ -232,32 +244,123 @@ cleanup()
 	if (rc)
 	    log_db_error("db_env_create failed during cleanup", rc);
 	else {
+	    muffle_error++;
 	    rc = dbenv->remove(dbenv, db_home, 0);
 	    if (rc)
-		syslog(LOG_WARNING, "BDB-%d: db_env->remove failed: %s",
+		syslog(LOG_DEBUG, "DEBUG: BDB-%d: db_env->remove failed: %s",
 		       rc, db_strerror(rc));
+	    muffle_error--;
 	}
     }
-    if (debug_me)
-	syslog(LOG_DEBUG, "Cleaned");
+    policy_cleanup();
 }
 
 static void
 fatal(const char *msg)
 {
-    int err = errno;;
+    int err = errno;
     cleanup();
     syslog(LOG_ERR, "fatal: %s: %s", msg, db_strerror(err));
     abort();
 }
 
-static void *
-xrealloc(void *ptr, size_t size)
+static void
+run_expiry()
 {
-    void *newptr = realloc(ptr, size);
-    if (newptr)
-	return newptr;
-    fatal("Out of memory in xrealloc");
+    DBC *dbcp;
+    int rc;
+    time_t now;
+    unsigned int count = 0;
+    if (time(&now) == (time_t)-1) {
+	syslog(LOG_ERR, "time failed during run_expiry");
+	return;
+    }
+    muffle_error++;
+    if (rc = db->cursor(db, 0, &dbcp, 0))
+	log_db_error("db->cursor failed during expiry run", rc);
+    else {
+	DBT key = { 0 };
+	while ((rc = dbcp->c_get(dbcp, &key, &dbdata, DB_NEXT | DB_RMW)) == 0) {
+	    time_t ref_time;
+	    double age_max, age;
+	    if (triplet_data.pass_count) {
+		ref_time = triplet_data.access_time;
+		age_max = pass_max_idle;
+	    }
+	    else {
+		ref_time = triplet_data.create_time;
+		age_max = bloc_max_idle;
+	    }
+	    age = difftime(now, ref_time);
+	    if (age > age_max) {
+		if (opt_verbose)
+		    syslog(LOG_INFO, "Expiring %s %s after %.0f seconds idle",
+			   key.data,
+			   triplet_data.pass_count ? "pass" : "block", age);
+		if (rc = dbcp->c_del(dbcp, 0))
+		    log_db_error("dbcp->c_del failed", rc);
+		else
+		    count++;
+	    }
+	}
+	if (rc != DB_NOTFOUND)
+	    log_db_error("dbcp->c_get failed", rc);
+	if (rc = dbcp->c_close(dbcp))
+	    log_db_error("dbcp->c_close failed", rc);
+    }
+    muffle_error--;
+    if (count)
+	syslog(LOG_NOTICE, "Expired %u triplets", count);
+}
+
+/* Write the result of calling ctime(3) on stdout without the trailing
+   newline. */
+static void
+write_ctime(const time_t *time)
+{
+    const char *s, *t;
+    s = ctime(time);
+    t = strchr(s, '\n');
+    if (t)
+	fwrite(s, 1, t - s, stdout);
+    else
+	fputs(s, stdout);
+}
+
+static void
+do_dump_triplets()
+{
+    DBC *dbcp;
+    int rc;
+    DBT key = { 0 },
+	data = { 0 };
+    rc = db->cursor(db, 0, &dbcp, 0);
+    if (rc)
+	fprintf(stderr, "DBD-%d: db->cursor failed: %s\n",
+		rc, db_strerror(rc));
+    else {
+	while ((rc = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0) {
+	    const char *s = key.data;
+	    const struct triplet_data *t = data.data;
+	    printf("%s\t", s);
+	    s += strlen(s) + 1;
+	    printf("%s\t", s);
+	    s += strlen(s) + 1;
+	    fwrite(s, 1, key.size - (s - (char *)key.data), stdout);
+	    putchar('\t');
+	    write_ctime(&t->create_time);
+	    putchar('\t');
+	    write_ctime(&t->access_time);
+	    printf("\t%lu\t%lu\n", t->block_count, t->pass_count);
+	}
+	if (rc != DB_NOTFOUND)
+	    fprintf(stderr, "DBD-%d: dbcp->c_get failed: %s\n",
+		    rc, db_strerror(rc));
+	rc = dbcp->c_close(dbcp);
+	if (rc)
+	    fprintf(stderr, "DBD-%d: dbcp->c_close failed: %s\n",
+		    rc, db_strerror(rc));
+    }
 }
 
 static void
@@ -310,7 +413,7 @@ build_triplet_key(const char *ip, const char *from, const char *to)
 	    buf += n;
 	    /* XXX the byte we are subsituting can only be smaller
 	       than the original so no additional memory is needed. */
-	    n = sprintf(buf, "%u", byte & ipv4_network_mask);
+	    n = sprintf(buf, "%lu", byte & ipv4_network_mask);
 	    buf += n;
 	    assert(buf - (char *)dbkey.data <= lenip);
 	}
@@ -349,102 +452,6 @@ build_data()
     triplet_data.pass_count = 0;
 }
 
-/**********************************************************************
- * See SMTPD_POLICY_README
- */
-
-static void
-safe_writes(int fd, const char *s)
-{
-    size_t len = strlen(s);
-    while (len) {
-	int n = write(fd, s, len);
-	if (n > 0)
-	    len -= n;
-	else if (n == 0)
-	    syslog(LOG_ERR, "Retrying due to empty write");
-	else if (errno == EINTR && debug_me)
-	    syslog(LOG_DEBUG, "Retrying write: %s", strerror(errno));
-    }
-}
-
-/* XXX Assumes there is an empty line marking the end of request */
-static char *
-find_attribute(const char *name)
-{
-    char *p;
-    size_t nlen;
-    nlen = strlen(name);
-    for (p = policy_request; *p != '\n'; p = strchr(p, '\n') + 1)
-	if (strncmp(name, p, nlen) == 0 && p[nlen] == '=')
-	    return p + nlen + 1;
-    return NULL;
-}
-
-static char *
-find_empty_line(const char *p, const char *const endp)
-{
-    while (p < endp && (p = memchr(p, '\n', endp - p))) {
-	p++;
-	if (p < endp && *p == '\n')
-	    return (char *) p + 1;
-    }
-    return NULL;
-}
-
-/* Find the end of request marker */
-static char *
-find_eor()
-{
-    return find_empty_line(policy_request,
-			   policy_request + policy_request_fill);
-}
-
-/* Forget the oldest policy request */
-static void
-forget_policy_request()
-{
-    const char *eor = find_eor();
-    if (eor) {
-	policy_request_fill -= eor - policy_request;
-	memmove(policy_request, eor, policy_request_fill);
-    }
-    else
-	policy_request_fill = 0;
-}
-
-/* Read in a new SMTPD access policy request */
-static const char *
-read_policy_request(int in)
-{
-    forget_policy_request();
-    while (! find_eor()) {
-	size_t wanted;
-	int n;
-	/* Make sure there is some room to read data */
-	if (policy_request_fill == policy_request_size) {
-	    if (policy_request_size)
-		policy_request_size *= 2;
-	    else
-		policy_request_size = BUFSIZ;
-	    if (debug_me)
-		syslog(LOG_DEBUG,
-		       "allocate %u bytes for request buffer",
-		       policy_request_size);
-	    policy_request = xrealloc(policy_request, policy_request_size);
-	}
-	wanted = policy_request_size - policy_request_fill;
-	n = read(in, policy_request + policy_request_fill, wanted);
-	if (n < 0)
-	    log_db_error("read_policy_request failed", n);
-	else if (n)
-	    policy_request_fill += n;
-	else
-	    return NULL;
-    }
-    return policy_request;
-}
-
 static void
 get_grey_data()
 {
@@ -456,8 +463,23 @@ get_grey_data()
 	log_db_error("get failed", rc);
 	fatal("Exiting");
     }
-    else
-	touch_data();
+    else {
+	time_t ref_time, now;
+	double age_max, age;
+	if (triplet_data.pass_count) {
+	    ref_time = triplet_data.access_time;
+	    age_max = pass_max_idle;
+	}
+	else {
+	    ref_time = triplet_data.create_time;
+	    age_max = bloc_max_idle;
+	}
+	age = difftime(now, ref_time);
+	if (age > age_max)
+	    build_data();
+	else
+	    touch_data();
+    }
 }
 
 static void
@@ -470,74 +492,65 @@ put_grey_data()
 }
 
 static void
-stats_expire()
+printf_action(const char *fmt, unsigned long delay)
 {
-    int rc;
-    DBT key;
-    DBT data;
-    memset(&key, 0, sizeof key);
-    memset(&data, 0, sizeof data);
-    if (triplet_data.pass_count) {
-	;
-    }
-    else {
-	unsigned long count;
-	key.data = STAT_BLOCKED;
-	key.size = strlen(STAT_BLOCKED);
-	data.data = &count;
-	data.ulen = sizeof(count);
-	data.flags = DB_DBT_USERMEM;
-	rc = statsdb->get(statsdb, NULL, &key, &data, 0);
-	if (rc == DB_NOTFOUND)
-	    count = 1;
-	else if (rc) {
-	    syslog(LOG_ERR, "BDB-%d: stats_expire block get: %s",
-		   rc, db_strerror(rc));
-	    return;
+    /* interpret the %x sequences */
+    while (*fmt) {
+	const char *t = strchr(fmt, '%');
+	if (t) {
+	    fwrite(fmt, sizeof *fmt, t - fmt, stdout);
+	    fmt = t + 1;
+	    switch (*fmt) {
+	    case 'd':
+		printf("%lu", delay);
+		break;
+	    case 'p':
+		fputs(delay == 1 ? singular_string : plural_string, stdout);
+		break;
+	    case 's':
+		putchar(' ');
+		break;
+	    case '%':
+		putchar('%');
+		break;
+	    case 0:
+		syslog(LOG_WARNING, "Invalid reject_action sequence %%");
+		break;
+	    default:
+		syslog(LOG_WARNING,
+		       "Invalid reject_action sequence %%%c", *fmt);
+		break;
+	    }
+	    if (*fmt)
+		fmt++;
+	}
+	else {
+	    fputs(fmt, stdout);
+	    break;
 	}
-	rc = statsdb->put(statsdb, NULL, &key, &data, 0);
-	if (rc)
-	    syslog(LOG_ERR, "BDB-%d: stats_expire block put: %s",
-		   rc, db_strerror(rc));
     }
 }
 
-static int
-triplet_expired_p(const struct triplet_data *data)
-{
-    unsigned int age = triplet_data.access_time - triplet_data.create_time;
-    return triplet_data.pass_count == 0 && age > bloc_max_idle
-	|| triplet_data.pass_count != 0 && age > pass_max_idle;
-}
-
 static void
 process_smtp_rcpt()
 {
-    unsigned long delay;
+    double delay;
     get_grey_data();
-    /* Expire records */
-    if (triplet_expired_p(&triplet_data)) {
-	if (debug_me)
-	    syslog(LOG_DEBUG, "expired record");
-	stats_expire();
-	triplet_data.create_time = triplet_data.access_time;
-    }
-    delay = triplet_data.access_time - triplet_data.create_time;
+    delay = difftime(triplet_data.access_time, triplet_data.create_time);
     /* Block inbound mail that is from a previously unknown (ip, from, to) triplet */
     if (delay < greylist_delay) {
-	delay = greylist_delay - delay;
 	triplet_data.block_count++;
-	printf("action=DEFER_IF_PERMIT Greylisted by " PACKAGE_STRING ", "
-	       "try again in %lu second%s."
-	       "  See http://projects.puremagic.com/greylisting/ for more information.\n",
-	       delay, delay == 1 ? "" : "s");
+	fputs(STR_ACTION, stdout);
+	printf_action(reject_action_fmt, greylist_delay - delay);
+	putchar('\n');
     }
     else if (triplet_data.pass_count++)
-	puts("action=DUNNO");
-    else
-	printf("action=PREPEND X-Greyfix: Greylisted by " PACKAGE_STRING
-	       " for %lu second%s\n",
-	       delay, delay == 1 ? "" : "s");
+	puts(STR_ACTION "DUNNO");
+    else {
+	fputs(STR_ACTION, stdout);
+	printf_action(greylisted_action_fmt, delay);
+	putchar('\n');
+    }
     put_grey_data();
 }
 
@@ -550,10 +563,16 @@ signal_handler(int signal)
     exit(-1);
 }
 
+static int
+loptp(const char *opt, const char *long_opt)
+{
+    return strcmp(opt, long_opt) == 0;
+}
+
 static int
 optp(const char *opt, const char *short_opt, const char *long_opt)
 {
-    return strcmp(opt, short_opt) == 0 || strcmp(opt, long_opt) == 0;
+    return strcmp(opt, short_opt) == 0 || loptp(opt, long_opt);
 }
 
 int
@@ -612,6 +631,19 @@ main(int argc, const char **argv)
 		fatal("Invalid argument to --pass-max-idle.  "
 		      "Integer value expected");
 	}
+	else if (optp(argv[i], "-r", "--reject-action")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --reject-action");
+	    reject_action_fmt = argv[i];
+	}
+	else if (optp(argv[i], "-G", "--greylisted-action")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --greylisted-action");
+	    greylisted_action_fmt = argv[i];
+	}
+	else if (loptp(argv[i], "--dump-triplets")) {
+	  dump_triplets = 1;
+	}
 	else {
 	    fprintf(stderr, "Unknown option \"%s\"\n", argv[i]);
 	    exit(EXIT_FAILURE);
@@ -621,8 +653,6 @@ main(int argc, const char **argv)
     ipv4_network_strip_bytes = suffix >> 3;
     ipv4_network_mask = 0xffU & ~0U << (suffix & 0x7U);
 
-    if (opt_verbose)
-        syslog(LOG_NOTICE, "daemon started");
 #ifdef SIGHUP
     signal(SIGHUP, signal_handler);
 #endif
@@ -653,12 +683,14 @@ main(int argc, const char **argv)
 	errno = rc;
 	fatal("initialization failure");
     }
-    while (read_policy_request(0)) {
+    if (dump_triplets)
+	do_dump_triplets();
+    else while (read_policy_request(0)) {
 	const char *protocol = 0, *state = 0, *ip, *from, *to;
 	if ((protocol = find_attribute("protocol_name"))
 	    && (state = find_attribute("protocol_state"))
-	    && (prefixp(protocol, str_smtp) || prefixp(protocol, str_esmtp))
-	    && prefixp(state, str_rcpt)
+	    && (prefixp(protocol, STR_SMTP) || prefixp(protocol, STR_ESMTP))
+	    && prefixp(state, STR_RCPT)
 	    && (ip = find_attribute("client_address"))
 	    && (from = find_attribute("sender"))
 	    && (to = find_attribute("recipient"))) {
@@ -666,7 +698,7 @@ main(int argc, const char **argv)
 	    process_smtp_rcpt();
 	}
 	else {
-	    puts("action=DUNNO");
+	    puts(STR_ACTION "DUNNO");
 	    if (debug_me) {
 		char *p = 0, *s = 0;
 		if (protocol) {
@@ -690,6 +722,7 @@ main(int argc, const char **argv)
 	putchar('\n');
 	fflush(stdout);
     }
+    run_expiry();
     cleanup();
     if (opt_verbose)
 	syslog(LOG_NOTICE, "daemon stopped");
